//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Class          : Rational
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   Objects constructed from the Rational class act like fractions
// (rational numbers have a sign, numerator, and denominator). This
// class implements some standard arithmetic and relational operations
// (as methods) on values of this class.
//
// Known Bugs (if any): Arithmetic overflows are possible;
//                      upgrade using BigInteger
//
// Future Plans   : Document via JavaDoc (Done: 8/10/04)
//
// Program History:
//   6/05/01: R. Pattis - Operational for 15-100
//   8/10/04: R. Pattis - Javadoc comments added (added to course library)
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


package edu.cmu.cs.pattis.cs151xx;



/** 
 * Objects constructed from the Rational class act like fractions
 *   (rational numbers have a sign, numerator, and denominator).
 * This class implements some standard arithemetic and relational
 *  operations (as methods) on values of this class.
 *
 * Object constructed from this class obey the following class
 * invariants
 *   <ul>
 *     <li>Zero is always stored as 0/1
 *     <li>Denominators are always positive
 *     <li>Numerators and Denominators are in lowest common terms
 *           (they have no common factors)
 *   </ul>
 * Note that objcets constructed from this class are immutable, so
 *   once these invariants are satisfied by the constructor, they
 *   always stay true.
*/
public class Rational {


	/** 
	 * Constructs a <code>Rational</code> object, specifying the
	 *   numerator and denominator of the fraction.
	 * These value are examined and put in a canonical form to ensure 
	 *   the following class invariants
	 *   <ul>
	 *     <li>Zero is always stored as 0/1
	 *     <li>Denominators are always positive
	 *     <li>Numerators and Denominators are in lowest common terms
	 *           (they have no common factors)
	 *   </ul>
	 *
	 * @param  numerator specifies the numerator of the fraction
	 * @param  denominator specifies the denominator of the fraction
	 *
	 * @throws IllegalArgumentException if <code>denominator</b> is 0
	*/
  public Rational (int numerator, int denominator)
    throws IllegalArgumentException
  {
    if (denominator == 0)
      throw new IllegalArgumentException("Exception in Rational Construtor; - denominator 0");

    //Ensure 0 is stored as 0/1
    if (numerator == 0)
      denominator = 1;
    
    //Ensure non-negative denominator; if a fraction is negative, its
    //  numerator is negative
    if (denominator < 0) {
      denominator = -denominator;
      numerator   = -numerator;
    }
    
    int common = gcd(numerator,denominator);  //or Rational.gcd(...)
    
    //Ensure that there are no common factors
    this.numerator   = numerator  /common;
    this.denominator = denominator/common;
  }



	/** 
	 * Constructs a <code>Rational</code> object, specifying the
	 *   numerator of the fraction (the denominator is one).
	 *
	 * @param  numerator specifies the numerator of the fraction
	*/
  public Rational (int numerator)
  {this(numerator, 1);}
  
  
  
	/** 
	 * Constructs a <code>Rational</code> object 0/1.
	*/
  public Rational ()
  {this(0, 1);}
  



	/** 
	 * Constructs a <code>Rational</code> object whose state is the
	 *   same as <code>other</code>.
	 *
	 * @param other specifies the fraction whose value to store into
	 *          the state of this object
	*/
  public Rational (Rational other)
  {this(other.numerator, other.denominator);}
  
  
  
  
	/** 
	 * Returns the numerator of the fraction.
	 *
	 * @return the numerator of the fraction
	*/
  public int getNumerator()
  {return this.numerator;}
  
  
  
	/** 
	 * Returns the denominator of the fraction.
	 *
	 * @return the denominator of the fraction
	*/
  public int getDenominator()
  {return this.denominator;}
  
  
  
	/** 
	 * Returns whether or not this fraction is equal to <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the equality check
	 *
	 * @return whether or not this fraction is equal to <code>other</code>
	*/
  public boolean equals (Rational other)
  {
    //Note: works because numbers are always stored in a canonical form
    return this.numerator  == other.numerator &&
           this.denominator == other.denominator;
   }
  

  
	/** 
	 * Returns the trichotomous comparison of this value and
	 *   <code>other</code>.
	 * More specifically:
	 * <ul>
	 *   <li>a negative value if <code>this</code> &lt <code>other</code>
	 *   <li>zero if <code>this</code> = <code>other</code>
	 *   <li>a positive value if <code>this</code> &gt <code>other</code>
	 * </ul>
	 * Example use, checking whether <code>f1</code> is less than or equal
	 *   to <code>f2</code>:<p>
	 *   <b><pre>  if (f1.compareTo(f2) <= 0)</pre></b>
	 *
	 * @param  other specifies the other fraction in the comparison
	 *
	 * @return the trichotomous comparison of this value and
	 *   <code>other</code>
	*/
  public int compareTo (Rational other)
  {
    int left  = this.numerator   * other.denominator;
    int right = this.denominator * other.numerator;
    
    //  a     c
    // --- ? --- = ad ? bc
    //  b     d
    
    return left-right;
    
    //Another way to implement this body is
    //if (left < right)
    //  return -1;
    //else if (left == right)
    //  return 0;
    //else return +1;
  }
  
  
  
	/** 
	 * Returns a new fraction that is the absolute value of this one.
	 *
	 * @return a new fraction that is the absolute value of this one.
	*/
  public Rational abs()
  {return new Rational(Math.abs(this.numerator),this.denominator);}
  
  
  
	/** 
	 * Returns a new fraction that is the negated value of this one.
	 *
	 * @return a new fraction that is the negated value of this one.
	*/
  public Rational negate()
  {return new Rational(-this.numerator,this.denominator);}
  
  
  
	/** 
	 * Returns a new fraction that is the reciprocal value of this one:
	 *   1/<code>this</code>.
	 *
	 * @return a new fraction that is the reciprocal value of this one:
	 *   1/<code>this</code>
	 *
	 * @throws IllegalStateException if this value is 0
	*/
  public Rational reciprocal()
    throws IllegalStateException
  {
    //In this version, check directly (see divide for another way)
    if (this.equals(ZERO))
      throw new IllegalStateException("Exception in Rational reciprocal: attempt to invert zero");

    return new Rational(this.denominator,this.numerator);
  }
  
  
  
	/** 
	 * Returns a new fraction that is the sum of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the sum
   *
	 * @return a new fraction that is the sum of this one and
	 *   <code>other</code>
	*/
  public Rational add (Rational other)
  { 
    int a = this.numerator;          //  a     c     ad + cb
    int b = this.denominator;        // --- + --- = ---------
    int c = other.numerator;         //  b     d        bd
    int d = other.denominator;       //
    
    return new Rational(a*d + c*b, b*d);
  }
  
  
  
	/** 
	 * Returns a new fraction that is the difference of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the difference
   *
	 * @return a new fraction that is the difference of this one and
	 *   <code>other</code>
	*/
  public Rational subtract (Rational other)
  { 
    int a = this.numerator;          //  a     c     ad - cb
    int b = this.denominator;        // --- - --- = ---------
    int c = other.numerator;         //  b     d        bd
    int d = other.denominator;       //
    
    return new Rational(a*d - c*b, b*d);
    
    //or write the following, which is compact but a bit slower
    //  this.add(other.negate())
  }
  
  
  
	/** 
	 * Returns a new fraction that is the product of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the product
   *
	 * @return a new fraction that is the product of this one and
	 *   <code>other</code>
	*/
  public Rational multiply (Rational other)
  { 
    int a = this.numerator;          //  a     c     ac
    int b = this.denominator;        // --- * --- = ----
    int c = other.numerator;         //  b     d     bd
    int d = other.denominator;       //
    
    return new Rational(a*c,b*d);
  }
  
  
  
	/** 
	 * Returns a new fraction that is the dividend of this one and
	 *   <code>other</code>.
	 *
	 * @param  other specifies the other fraction in the division
   *
	 * @return a new fraction that is the dividend of this one and
	 *   <code>other</code>
	 *
	 * @throws IllegalArgumentException if <code>other</b> value is 0
	*/
  public Rational divide (Rational other)
    throws IllegalArgumentException
  { 
    int a = this.numerator;          //  a     c     ad
    int b = this.denominator;        // --- / --- = ----
    int c = other.numerator;         //  b     d     bc
    int d = other.denominator;       //
    
    return new Rational(a*d,b*c);

    //or write the following, which is compact but a bit slower
    //  this.multiply(other.reciprocal())
  }
  


	/** 
	 * Returns a <code>String<code> representation of the state of this 
	 *  fraction: numerator/denominator.
	 *
	 * @return a <code>String<code> representation of the state of this 
	 *  fraction: numerator/denominator
	*/
  public String toString()
  {return this.numerator + "/" + this.denominator;}
  
  

	/** 
	 * Returns a <code>String<code> representation of the state of this 
	 *  fraction, as a decimal value approximating the fraction.
	 * For example <code>(new Rational(1,3)).toDecimalString(20)</code>
	 *   returns <code>".33333333333333333333"</code>;
	 *  <code>(new Rational(1,7)).toDecimalString(13)</code>
	 *   returns <code>".1428571428571"</code>.
	 *
	 * @param  decimalPlaces specifies the number of decimal places to use
	 *           in the approximation
   *
	 * @return a <code>String<code> representation of the state of this 
	 *  fraction, as a decimal value approximating the fraction
	*/
  public String toDecimalString(int decimalPlaces)
  {
    //This method performs the equivalent of infinite precsison
    //  long division!
    //If you are curious, hand simulate it on a few examples

    String answer = new String("");
    
    int num   = this.numerator;
    int denom = this.denominator;
    
    //Handle negative numbers
    if (num < 0) {
      num = -num;
      answer = answer + "-";
    }
    
    //Handle integer part
    if (num >= denom) {
      answer = answer + num/denom;  //Catenate
      num    = num%denom;
    }

    //Handle decimal part
    answer = answer + ".";
    num   *= 10;
    for (int i=1; i<=decimalPlaces; i++) {
	    answer += num/denom;          //Catenate
	    num  = num%denom;
      num *= 10;
	  }
      
    return answer;
  }
  
  
  
	/** 
	 * Returns a new fraction after prompting the user for its numerator
	 *   and denominator.
	 * If an illegal value is entered, the user is reprompted
	 * For example, if we write <code>Rational.prompt("Enter height")</code>
	 *   an interaction on the console screen might look like
	 * <b><pre>  Enter Height
	 *    Enter numerator  : 1
	 *    Enter denominator: 3</pre></b>
	 *
	 * @param message specifies the message with which to prompt the user
   *
	 * @return a new fraction after prompting the user for its numerator
	 *   and denominator.
	*/
  public static Rational prompt(String message)
  {
    System.out.println(message);
    for (;;)
      try{
        int numerator   = Prompt.forInt("  Enter numerator  ");
        int denominator = Prompt.forInt("  Enter denominator");
        return new Rational(numerator,denominator);
      }catch (IllegalArgumentException iae)
        {System.out.println("Illegal rational entered; please try again");}
  }



	/** 
	 * Returns the greatest common divisor of its parameters.
	 *
	 * @param x one value to compute the greatest common divisor of
	 * @param y one value to compute the greatest common divisor of
	  the message with which to prompt the user
   *
	 * @return the greatest common divisor of its parameters
	*/
  //For simplifying rationals in the main constructor
  private static int gcd (int x, int y)
  {
    x = Math.abs(x);
    y = Math.abs(y);
    
    for (;;) {
      if (x == 0 || y == 0)
         return x+y;
         
      if (x > y)
        x %= y;
      else
        y %= x;
    }
  }



	/** 
	 * Stores the numerator of the fraction.
	 * It may be positive or negative.
	 * It will have no common factors with the denominator.
	*/
  private int numerator;
  
	/** 
	 * Stores the denominator of the fraction.
	 * It is always positive, and stores 1 if the numerator is 0.
	 * It will have no common factors with the numerator.
	*/
  private int denominator;
  
  
	/** 
	 * Stores the fraction 0/1.
	*/
  public static final Rational ZERO = new Rational();
  
  
  /** 
	 * Stores the fraction 1/1.
	*/
  public static final Rational ONE  = new Rational(1);

}
